/**
  Copyright 2002 Bruce Spencer 
*/

import java.util.*;



/**
   <code>DefiniteClause</code> objects consist of an array of arrays of
   size 2, where each inner array consists of, in the first
   position, either an index into a symbol table or variable
   table and, in the second position, a length.  The symbol table
   is a static array of strings stored in the TPTPFileParser
   class. The variable table is not stored with the clause, but
   rather with a Unifier object that is created at various times,
   such as during resolution and factoring steps. But the indices
   into that table are stored in the clause as negative integers.
   (Note that the negative numbers are indexed from -1 but the
   variable table in Unifier is index from 0.)
   
   DefiniteClauses are stored sequences as of terms which represent the
   literals in the clause.  DefiniteClauses, and indeed all terms, are
   stored as "flat" with lengths, which means that the term
   structure, telling which terms are arguments to which, is
   encoded with the lengths: those second items in the arrays of
   size 2.  For instance if we are representing p(a, q(b, X)) the
   symbols array is {{^p,5},{^a,1},{^q,3},{^b,1},{-1,1}}.  Here
   ^p is used to represent that index into the symbol table where
   "p" is stored and the -1 indicates the first variable in the
   variable table.
   

   @author Bruce Spencer <bspencer@unb.ca>
   @version $Id: DefiniteClause.java,v 1.5 2002/07/15 18:29:48 userid Exp $
   
   Created: Wed Jul 26 20:57:59 2000 */

public class DefiniteClause{

  
  /**
   <code>clauseCount</code> counts the number of clauses created so far
   */
  public static int clauseCount=0;


  /**
    <code>clauseNumber</code> is the number of the clause as it
   is created, or if  renumbered is true, it is the number of
   the clause in the proof
   */
    public int clauseNumber; //takes the current clauseCount

  /**
   <code>varCount</code> counts the number of variables in this
   clause so the unifier will know how many replacements the
   substitution will need.  */
  public int varCount;


  /**
   <code>atomCount</code> counts the number of literals in the clause
   */
  public int atomCount;


  /**
     The variable <code>symbolTable</code> refers to the
     SymbolTable where the printNames for the symbols in this
     DefiniteClause are found.  THis is a public field because we
     intend to use it to test the reference for equality with
     other symbolTable references and to know how to print this
     clause.  */
  public SymbolTable symbolTable;

  /**
    <code>symbols</code> refers to the symbols, variables and
    the lengths of the subterms in this clause, as described above.
   */
  public int[][] symbols;
  
  /**
     Creates a new <code>DefiniteClause</code> instance for an initial clause

     @param syms - a <code>Vector</code> value, containing the symbols
     @param vars - a <code>Hashtable</code> value, containing the variable names
     @param atomCount - an <code>int</code> value
     @param symbolTable - a <code>SymbolTable</code> value, so we
     can trace the pedigree of this clause and also to know where
     the print names for it are.  */
  DefiniteClause(Vector syms, Hashtable vars, int atomCount, SymbolTable symbolTable){
    loadVariableNames(vars);
    this.atomCount = atomCount;
    this.symbolTable = symbolTable;
    loadSymbols(syms);
    clauseNumber = clauseCount;
    clauseCount++;
  }

  String[] variableNames;
  boolean hasVariableNames;

  private void loadVariableNames(Vector vars){
    varCount = vars.size();
    hasVariableNames = true;
    variableNames = new String[varCount];
    Iterator it = vars.iterator();
    int vPos = 0;
    while(it.hasNext()){
      String varName = (String) it.next();
      variableNames[vPos] = varName;
      vPos++;
    }
  }

  private void loadVariableNames(Hashtable vars){
    varCount = vars.size();
    hasVariableNames = true;
    variableNames = new String[varCount];
    Iterator it = vars.entrySet().iterator();
    while(it.hasNext()){
      Map.Entry me = (Map.Entry) it.next();
      int pos = ((Integer) me.getValue()).intValue();
      String name = (String) me.getKey();
      variableNames[pos-1] = name;//sub 1 because vars are indexed from 1
    }
  }

  /**
     Creates a new <code>DefiniteClause</code> instance for an initial clause

     @param syms - a <code>Vector</code> value, containing the symbols
     @param varCount - an <code>int</code> value
     @param atomCount - an <code>int</code> value
     @param symbolTable - a <code>SymbolTable</code> value, so we
     can trace the pedigree of this clause and also to know where
     the print names for it are.  */
  DefiniteClause(Vector syms, int varCount, int atomCount, SymbolTable symbolTable){
    this.varCount = varCount;
    this.atomCount = atomCount;
    this.symbolTable = symbolTable;
    hasVariableNames = false;
    loadSymbols(syms);
    clauseNumber = clauseCount;
    clauseCount++;
  }

  /**
     <code>loadSymbols</code> loads the symbols array for this
     clause from the Vector v containing an array values of
     length 2.  In either case the symbol consists of a either an
     index into the symbol table or a variable (which is a
     negative numbered index)

     @param v - a <code>Vector</code> value */
  private void loadSymbols(Vector v){
    symbols = new int[v.size()][2];
    Iterator it = v.iterator();
    int i = 0;
    while(it.hasNext()){
      Object next = it.next();
      int[] se = (int[]) next;
      symbols[i] = se;
      i++;
    }
  }


  /**
     <code>dump</code> displays the clause in raw mode of System.out

  */
//    void dump(){
//      for(int i = 0; i < symbols.length; i++){
//        if(symbols[i][0]<0){
//  	System.out.println(i+":"+"V"+(-symbols[i][0]));
//        }
//        else if(symbols[i][0] >= symbolTable.size()){
//  	System.out.println(i+":"+
//  			   "$"+i+":"+
//  			   symbols[i][1]);
//        }
//        else {
//  	System.out.println(i+":"+
//  			   symbolTable.symbolTableString(symbols[i][0])+":"+
//  			   symbols[i][1]);
//        }
//      }
//    }

  public String toString(){
    StringBuffer b = new StringBuffer();
    int endOfHead = symbols[0][1];
    bufferAppend(b, 0, endOfHead);
    if(endOfHead < symbols.length){
      b.append(":-");
      bufferAppend(b, endOfHead, symbols.length);
    }
    b.append(".");
    
    return b.toString();
  }

  /**
     <code>println</code> prints the clause in a standard format
     followed by a new line character

  */
  void println(){
    System.out.println(toString());
  }

  /**
     <code>println</code> prints the clause in a standard format
  */
  void print(){
    System.out.println(toString());
  }

  /**
     <code>bufferAppend</code> is the auxilliary function for printing a clause

     @param b - a <code>StringBuffer</value>
     @param from - an <code>int</code> value telling at which
     * symbol to start printing
     @param to - an <code>int</code> value telling which symbol
     * is ONE PAST the point where we stop printing
  */
  void bufferAppend(StringBuffer b, int from, int to){
    while(from<to){
      int symb = symbols[from][0];
      if(symb<0){
	if(hasVariableNames){
	  b.append(variableNames[-symb-1]);
	}
	else {
	  b.append("V");
	  b.append(-symb-1);
	}
      }
      else if(symb >= symbolTable.size()){
	b.append("$"+symb);
      }
      else {
	String s = symbolTable.
	  symbolTableString(symb);

	  if (Character.isUpperCase(s.charAt(0)) ||
	      ParserBasic.containsDelimiters(s)){
	    //if starts with upper case or contains delimiters
	    //print with quotes
	    b.append('\'');
	    b.append(s);
	    b.append('\'');
	  }
	  else{
	    b.append(s);
	  }

	int len = symbols[from][1];
	if(len>1){
	  b.append("(");
	  bufferAppend(b, from+1, from + len);
	  b.append(")");
	  from+=len-1;
	}
      }
      from++;
      if(from<to) b.append(",");
    }
  }

  String literalIndexToString(int from, int to){
    StringBuffer sb = new StringBuffer();
    bufferAppend(sb, from, to);
    return sb.toString();
  }

}

